﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.ComponentModel;
using System.Windows.Media;
using Infragistics.Olap;
using Infragistics.Olap.Data;
using Infragistics.Olap.Data.Xmla.SqlAnalysisServices2008;
using Infragistics.Olap.Xmla;
using MeasureGroup = Infragistics.Olap.Data.Base.MeasureGroup;
using Level = OlapAxisWithXMLA.Level;
using System.Windows.Input;

namespace OlapAxisWithXMLA
{
    public class AdomdDataProvider : XmlaDataProvider, INotifyPropertyChanged
    {
        private readonly AdomdConnection _connection;
        private bool _isFirstRun = true;
        /// <summary>
        /// 
        /// </summary>
        /// <param name="connection"></param>
        public AdomdDataProvider(IOlapConnection connection)
            : base(connection)
        {
            _connection = connection as AdomdConnection;
        }

        #region Properties

        /// <summary>
        /// 
        /// </summary>
        public override IConnectionSettings ConnectionSettings
        {
            get
            {
                return _connection.ConnectionSettings;
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public string DatabaseName { get; set; }

        #endregion // Properties

        /// <summary>
        /// 
        /// </summary>
        public override void LoadSchemaAsync()
        {
            BackgroundWorker w = new BackgroundWorker();
            w.RunWorkerCompleted += delegate
            {
                Mouse.OverrideCursor = Cursors.Wait;

                var schema = new OlapSchema();
                AddDatabase(schema, new Database(DatabaseName, DatabaseName, DatabaseName, DatabaseName));
                OnLoadSchemaCompleted(new LoadSchemaCompletedEventArgs(schema, null, false, schema));
                Mouse.OverrideCursor = null;
            };
            w.RunWorkerAsync();
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="catalogName"></param>
        public override void LoadCubesAsync(string catalogName)
        {
            var cubeCollection = ApplicationServices.LoadCubes();

            var schema =
            new OlapSchema();

            foreach (var cube in cubeCollection)
            {
                if (cube.CubeType == "Cube")
                {
                    AddCube(
                        schema,
                        new Cube(
                            cube.CubeName,
                            cube.Caption,
                            cube.Description,
                            CubeType.Cube,
                            cube.CatalogName,
                            cube.LastDataUpdate,
                            cube.LastDataUpdate));
                }
            }

            OnLoadCubesCompleted(new LoadCubesCompletedEventArgs(schema, null, false, schema));

        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="cube"></param>
        public override void LoadDimensionsAsync(ICube cube)
        {
            var dimensions = ApplicationServices.LoadCubeDimensions(cube.Name);

            foreach (var dimension in dimensions)
            {
                var dimensionType =
                    (DimensionType)Enum.Parse(typeof(DimensionType), dimension.DimensionType, true);

                var currentDimension = new Dimension(
                    dimension.Name,
                    dimension.UniqueName,
                    AdjustNameForCaption(dimension.Caption),
                    dimension.Caption,
                    dimensionType);

                foreach (var adomdHierarchy in dimension.Hierarchies)
                {
                    if (ValidDimensionAttribute(adomdHierarchy.Name))
                    {
                        var hierarchy =
                            new Hierarchy(
                                adomdHierarchy.Name,
                                adomdHierarchy.UniqueName,
                                AdjustNameForCaption(adomdHierarchy.Caption),
                                adomdHierarchy.Description,
                                adomdHierarchy.DefaultMember,
                                adomdHierarchy.DefaultMember,
                                currentDimension,
                                string.Empty,
                                HierarchyOrigin.SystemEnabled);

                        foreach (AdomdLevel adomdLevel in adomdHierarchy.Levels)
                        {
                          
                            Level level = new Level(
                            adomdLevel.Name,
                            adomdLevel.Caption,
                            adomdLevel.UniqueName,
                            adomdLevel.Description,
                            (int)adomdLevel.Number,
                            hierarchy,
                            (int)adomdLevel.MembersCount);

                            AddLevel(hierarchy, level);
                        }

                        AddHierarchy(currentDimension, hierarchy);
                    }
                }

                AddDimension((Cube)cube, currentDimension);
            }

            OnLoadDimensionsCompleted(new AsyncCompletedEventArgs(null, false, cube));
        }

        private static bool ValidDimensionAttribute(string name)
        {
            var nameLower = name.ToLower();
            // don't show any sids in the UI
            if (name.EndsWith("Sid") || name.EndsWith("SID"))
            {
                return false;
            }
            // don't show any unique fields in the UI...it is not useful there and clutters
            else if (name.EndsWith("Unique") || name.EndsWith("UNIQUE"))
            {
                return false;
            }
            // remove location specific values...Infragistics does not work well with such specific things
            // as it overloads the control and hangs the app
            else if (nameLower == "latitude"
                || nameLower == "longitude"
                || nameLower == "grossarea"
                || nameLower == "subarea2code"
                || nameLower == "crestacode")
            {
                return false;
            }
            return true;
        }

        private static string AdjustNameForCaption(string name)
        {
            return System.Text.RegularExpressions.Regex.Replace(name, "([a-z])([A-Z])", "$1 $2");
            /*switch (name.ToLower())
            {
                case "crestaname":
                    return "CRESTA Name";
                case "constructioncategory":
                    return "Construction Category";
                case "airconstruction":
                    return "Construction Type";
                case "occupancycategory":
                    return "Occupancy Category";
                case "airoccupancy":
                    return "Occupancy Type";
                case "perilsetcode":
                    return "Peril";
                case "geomatchlevel":
                    return "Geo Match Level";
                case "grossarea":
                    return "Gross Area";
                case "buildingvaluegrossarea":
                    return "Building Value Gross Area";
                case "countryname":
                    return "Country";
                case "countrycode":
                    return "Country Code";
                case "areaname":
                    return "Area";
                case "areacode":
                    return "Area Code";
                case "subareaname":
                    return "Subarea";
                case "subareacode":
                    return "Subarea Code";
                case "postalname":
                    return "Postal";
                case "postalcode":
                    return "Postal Code";
                case "contractid":
                    return "Policy ID";
                case "status":
                    return "Policy Status";
                case "lob":
                    return "Line of Business";
                case "inceptiondate":
                    return "Inception Date";
                case "expirationdate":
                    return "Expiration Date";
                case "insuredname":
                    return "Insured Name";
                case "policyform":
                    return "Policy Form";
            }

            if (name.ToLower().EndsWith("count"))
                return "Location Count";
            else if (name.ToLower().StartsWith("ReplacementValueA"))
                name = name.Replace("ReplacementValueA", "Replacement Value Building");
            else if (name.ToLower().StartsWith("ReplacementValueB"))
                name = name.Replace("ReplacementValueA", "Replacement Value Other");
            else if (name.ToLower().StartsWith("ReplacementValueC"))
                name = name.Replace("ReplacementValueA", "Replacement Value Contents");
            else if (name.ToLower().StartsWith("ReplacementValueD"))
                name = name.Replace("ReplacementValueA", "Replacement Value Time");

            name = name.Replace("AIR", "").Replace("DISTANCE", "Distance").Replace("SID", "SID ");

            var retVal = "";
            var previousCharacter = ' ';

            foreach (var character in name)
            {
                if (retVal.Length > 0)
                {
                    if (character.ToString() == character.ToString().ToUpper()
                        && (character == '_' || previousCharacter.ToString() != previousCharacter.ToString().ToUpper()))
                    {
                        retVal += " ";
                    }
                }
                if (character != '_')
                    retVal += character;

                previousCharacter = character;
            }

            if (retVal.ToLower().EndsWith(" band"))
            {
                retVal = retVal.Substring(0, retVal.Length - 5);
            }

            return retVal;*/
            // for now don't adjust...we are doing it in the UI
            //return name;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="cube"></param>
        public override void LoadMeasuresAsync(ICube cube)
        {
            var measures = ApplicationServices.LoadCubeMeasures(cube.Name);

            foreach (var measure in measures.GroupBy(m => FixMeasureIdentity(m.Name)).Select(g => g.First()))
            {

                var aggregator =
                    (AggregatorType)Enum.Parse(typeof(AggregatorType), measure.Aggregator, true);

                AddMeasure(
                    (Cube)cube,
                    new Measure(measure.Name, measure.UniqueName, FixMeasureIdentity(measure.Caption),
                        measure.Caption, measure.DefaultFormatString, measure.MeasureGroupName, aggregator));
            }

            OnLoadMeasuresCompleted(new AsyncCompletedEventArgs(null, false, cube));
        }

        private static string FixMeasureIdentity(string id)
        {
            //if (id.ToLower().EndsWith("sum") || id.ToLower().EndsWith("min") || id.ToLower().EndsWith("max"))
            //{
            //    return AdjustNameForCaption(id.Substring(0, id.Length - 3));
            //}
            return AdjustNameForCaption(id);
            // for now don't adjust...we are doing it in the UI
            //return id;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="cube"></param>
        /// <param name="filterMember"></param>
        public override void LoadMembersAsync(ICube cube, IFilterMember filterMember)
        {

            var members = ApplicationServices.LoadCubeFilterMemberMembers(cube.Name, filterMember.Member.UniqueName, filterMember.Member.ParentLevel.Depth + 1);

            var prentMember = filterMember.Member as Member;
            ClearMemberChildren(prentMember);

            foreach (var member in members)
            {
                MemberViewModel currentMember = new MemberViewModel(
                    member.Name,
                    member.UniqueName,
                    member.Caption,
                    member.ParentUniqueName,
                    member.LevelUniqueName,
                    (int)member.LevelNumber,
                    filterMember.Member.ParentLevel.ParentHierarchy.Levels[(int)member.LevelNumber],
                    member.ChildrenCardinality,
                    false,
                    false,
                    false);

                PopulateAIRPropertiesOnMember(member, currentMember);

                // infragistics tries to order members alphabetically, so we need to override
                // that in order to keep the order in the cube.  we store the members in the
                // correct order in the cube already so no need to reorder or try to guess
                // what the order is
                currentMember.OrderByKey = members.IndexOf(member).ToString().PadLeft(10, '0');
                AddMember(prentMember, currentMember);

            }
            OnLoadMembersCompleted(new AsyncCompletedEventArgs(null, false, filterMember));
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="cube"></param>
        /// <param name="level"></param>
        public override void LoadMembersAsync(ICube cube, ILevel level)
        {
            if (!_isFirstRun)
            {
                var members = ApplicationServices.LoadCubeLevelMembers(cube.Name, level.UniqueName);

                ClearMemberChildren(level as Level);

                foreach (var member in members)
                {
                    MemberViewModel currentMember;
                    currentMember = new MemberViewModel(
                        member.Name,
                        member.UniqueName,
                        member.Caption,
                        member.ParentUniqueName,
                        member.LevelUniqueName,
                        (int)member.LevelNumber,
                        level.ParentHierarchy.Levels[(int)member.LevelNumber],
                        member.ChildrenCardinality,
                        false,
                        false,
                        false);

                    PopulateAIRPropertiesOnMember(member, currentMember);

                    AddMember(level as Level, currentMember);
                }
            }
            _isFirstRun = false;
            OnLoadMembersCompleted(new AsyncCompletedEventArgs(null, false, level));
        }

        private static void PopulateAIRPropertiesOnMember(AdomdMember sourceMember, MemberViewModel destinationMember)
        {
            if (sourceMember.RGB != null)
            {
                //var colorValues = sourceMember.RGB.Split(',');
                // store additional properties retrieved from database
                destinationMember.Fill = ToDrawingColor(sourceMember.RGB);
                //colorValues.Length == 4
                //                     ? Color.FromArgb(
                //                         Convert.ToByte(colorValues[0]), Convert.ToByte(colorValues[1]),
                //                         Convert.ToByte(colorValues[2]), Convert.ToByte(colorValues[3]))
                //                     : colorValues.Length == 3
                //                           ? Color.FromArgb(
                //                               255, Convert.ToByte(colorValues[0]),
                //                               Convert.ToByte(colorValues[1]), Convert.ToByte(colorValues[2]))
                //                           : Colors.Red;
            }
            else
            {
                destinationMember.Fill = Colors.Red;
            }
        }
        private static Color ToDrawingColor(ColorInfo colorInfo)
        {
            return Color.FromArgb(colorInfo.A, colorInfo.R, colorInfo.G, colorInfo.B);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="cube"></param>
        /// <param name="hierarchy"></param>
        public override void LoadMembersAsync(ICube cube, IHierarchy hierarchy)
        {
            var members = ApplicationServices.LoadCubeMembers(cube.Name, hierarchy.UniqueName);

            for (var i = 0; i < hierarchy.Levels.Count; i++)
            {
                var level = (Level)hierarchy.Levels[i];
                ResolveLoadedMembers(members, level);
            }

            OnLoadMembersCompleted(new AsyncCompletedEventArgs(null, false, hierarchy));

        }

        private static void ResolveLoadedMembers(IEnumerable<AdomdMember> allElements, Infragistics.Olap.Data.Base.Level parenLevel)
        {
            var levelMembers = new List<Member>();
            foreach (var sourceMember in allElements)
            {
                if (sourceMember.LevelNumber == (uint)parenLevel.Depth)
                {
                    MemberViewModel levelMember;

                    levelMember = new MemberViewModel(
                        sourceMember.Name,
                        sourceMember.UniqueName,
                        sourceMember.Caption,
                        sourceMember.ParentUniqueName,
                        sourceMember.LevelUniqueName,
                        (int)sourceMember.LevelNumber,
                        parenLevel,
                        sourceMember.ChildrenCardinality,
                        false,
                        false,
                        false);

                    PopulateAIRPropertiesOnMember(sourceMember, levelMember);

                    levelMembers.Add(levelMember);
                    AddMember(parenLevel, levelMember);
                }
            }

            if (parenLevel.Depth > 0)
            {
                var upperLevel = parenLevel.ParentHierarchy.Levels[parenLevel.Depth - 1];
                if (upperLevel != null)
                {
                    foreach (var levelMember in levelMembers)
                    {
                        var member = levelMember;
                        var parentMember = (Member)upperLevel.Members.Where(m => m.UniqueName == member.ParentUniqueName).FirstOrDefault();
                        if (parentMember != null)
                        {
                            AddMember(parentMember, levelMember);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="cube"></param>
        public override void LoadMeasureGroupsAsync(ICube cube)
        {

            var measureGroups = ApplicationServices.LoadCubeMeasureGroups(cube.Name);

            if (measureGroups.Count > 0)
            {
                AddMeasureGroup(
                    (Cube)cube, MeasureGroup.AllMeasureGroup);
            }

            foreach (var measureGroup in measureGroups)
            {
                AddMeasureGroup(
                    (Cube)cube,
                    new MeasureGroup(measureGroup, measureGroup, measureGroup));
            }

            OnLoadMeasureGroupsCompleted(new AsyncCompletedEventArgs(null, false, cube));
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="cube"></param>
        public override void LoadMeasureGroupDimensionsAsync(ICube cube)
        {
            var measureGroupDimensions = ApplicationServices.LoadCubeMeasureGroupDimensions(cube.Name, ((Cube)cube).CatalogName);
            foreach (var keyValuePair in measureGroupDimensions)
            {
                foreach (var dimensionName in keyValuePair.Value)
                {
                    AddMeasureGroupDimensions((Cube)cube, keyValuePair.Key, dimensionName);
                }
            }

            OnLoadMeasureGroupDimensionsCompleted(new AsyncCompletedEventArgs(null, false, cube));
        }

        /// <summary>
        /// 
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        /// <summary>
        /// 
        /// </summary>
        /// <param name="p"></param>
        protected void OnPropertyChanged(string p)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(p));
        }
    }
}
